#include <stdio.h>
#include <string.h>
#include <io.h>
#include <string.h>
#include <stdlib.h>
#include <windows.h>
#include "tag.h"

#define COMPRESSION_LEVEL_FAST          1000
#define COMPRESSION_LEVEL_NORMAL        2000
#define COMPRESSION_LEVEL_HIGH          3000
#define COMPRESSION_LEVEL_EXTRA_HIGH    4000
#define COMPRESSION_LEVEL_INSANE        5000

#define MAC_FORMAT_FLAG_8_BIT                 1    // is 8-bit
#define MAC_FORMAT_FLAG_CRC                   2    // uses the new CRC32 error detection
#define MAC_FORMAT_FLAG_HAS_PEAK_LEVEL        4    // unsigned __int32 Peak_Level after the header
#define MAC_FORMAT_FLAG_24_BIT                8    // is 24-bit
#define MAC_FORMAT_FLAG_HAS_SEEK_ELEMENTS    16    // has the number of seek elements after the peak level
#define MAC_FORMAT_FLAG_CREATE_WAV_HEADER    32    // create the wave header on decompression (not stored)

typedef struct {
    char cID[4];                             // should equal 'MAC '
    unsigned __int16 nVersion;               // version number * 1000 (3.81 = 3810)
} APE_HEADER_ALL;

typedef struct {
    char cID[4];                             // should equal 'MAC '
    unsigned __int16 nVersion;               // version number * 1000 (3.81 = 3810)
    unsigned __int16 nCompressionLevel;      // the compression level
    unsigned __int16 nFormatFlags;           // any format flags (for future use)
    unsigned __int16 nChannels;              // the number of channels (1 or 2)
    unsigned __int32 nSampleRate;            // the sample rate (typically 44100)
    unsigned __int32 nHeaderBytes;           // the bytes after the MAC header that compose the WAV header
    unsigned __int32 nTerminatingBytes;      // the bytes after that raw data (for extended info)
    unsigned __int32 nTotalFrames;           // the number of frames in the file
    unsigned __int32 nFinalFrameBlocks;      // the number of samples in the final frame
} APE_HEADER_OLD;

typedef struct {
    char cID[4];                             // should equal 'MAC '
    unsigned __int16 nVersion;               // version number * 1000 (3.81 = 3810)
    unsigned __int16 padded;                 // padding/reserved (always empty)
    unsigned __int32 nDescriptorBytes;       // the number of descriptor bytes (allows later expansion of this header)
    unsigned __int32 nHeaderBytes;           // the number of header APE_HEADER bytes
    unsigned __int32 nSeekTableBytes;        // the number of bytes of the seek table
    unsigned __int32 nHeaderDataBytes;       // the number of header data bytes (from original file)
    unsigned __int32 nAPEFrameDataBytes;     // the number of bytes of APE frame data
    unsigned __int32 nAPEFrameDataBytesHigh; // the high order number of APE frame data bytes
    unsigned __int32 nTerminatingDataBytes;  // the terminating data of the file (not including tag data)
    char cFileMD5[16];                       // the MD5 hash of the file
    unsigned __int16 nCompressionLevel;      // the compression level
    unsigned __int16 nFormatFlags;           // any format flags (for future use)
    unsigned __int32 nBlocksPerFrame;        // the number of audio blocks in one frame
    unsigned __int32 nFinalFrameBlocks;      // the number of audio blocks in the final frame
    unsigned __int32 nTotalFrames;           // the number of frames in the file
    unsigned __int16 nBitsPerSample;         // the bits per sample (typically 16)
    unsigned __int16 nChannels;              // the number of channels (1 or 2)
    unsigned __int32 nSampleRate;            // the sample rate (typically 44100)
} APE_HEADER_NEW;

char* compressionLevel ( int compression )
{
    switch ( compression ) {
        case COMPRESSION_LEVEL_FAST       : return "Fast"; break;
        case COMPRESSION_LEVEL_NORMAL     : return "Normal"; break;
        case COMPRESSION_LEVEL_HIGH       : return "High"; break;
        case COMPRESSION_LEVEL_EXTRA_HIGH : return "Extra High"; break;
        case COMPRESSION_LEVEL_INSANE     : return "Insane"; break;
        default                           : return "Unknown"; break;
    }
}

int ReadFileInfoAPE ( FileInfo* Info )
{
    FILE*               fp;
    APE_HEADER_ALL      header_all;
    APE_HEADER_NEW      header_new;
    APE_HEADER_OLD      header_old;
    unsigned long       nBlocksPerFrame;
    int                 nBitsPerSample;
    unsigned __int64    samples;
    char*               comp;

    if ( (fp = fopen ( Info->Filename, "rb" )) == NULL )
        return 1;

    if ( (fread ( &header_all, 1, sizeof (header_all), fp )) != sizeof (header_all) ) {
        fclose ( fp );
        return 1;
    }

    if ( memcmp ( header_all.cID, "MAC ", 4 ) != 0 ) {
        fclose ( fp );
        return 1;
    }

    rewind ( fp );

    if ( header_all.nVersion >= 3980 ) {
        if ( (fread ( &header_new, 1, sizeof (header_new), fp )) != sizeof (header_new) ) {
            fclose ( fp );
            return 1;
        }
    } else {
        if ( (fread ( &header_old, 1, sizeof (header_old), fp )) != sizeof (header_old) ) {
            fclose ( fp );
            return 1;
        }
    }

    fclose ( fp );

    if ( header_all.nVersion >= 3980 ) {
        samples = (header_new.nTotalFrames == 0) ? 0 : ((header_new.nTotalFrames -  1) * header_new.nBlocksPerFrame) + header_new.nFinalFrameBlocks;
        Info->Details.Duration      = (header_new.nSampleRate > 0)  ?  ((double)((signed __int64)samples) / (double)header_new.nSampleRate)  :  0.;
        Info->Details.Samples       = (__int64)samples;
        Info->Details.SampleRate    = (long)header_new.nSampleRate;
        Info->Details.Channels      = (long)header_new.nChannels;
        Info->Details.BitRate       = (Info->Details.Duration > 0)  ?  (long)((signed __int64)Info->FileSize * 8 / Info->Details.Duration + 500) / 1000  :  0;
        comp = compressionLevel( header_new.nCompressionLevel );
    } else {
        nBlocksPerFrame = ((header_all.nVersion >= 3900) || ((header_all.nVersion >= 3800) && (header_old.nCompressionLevel == COMPRESSION_LEVEL_EXTRA_HIGH))) ? 73728 : 9216;
        if ((header_all.nVersion >= 3950)) nBlocksPerFrame = 73728 * 4;
        nBitsPerSample = (header_old.nFormatFlags & MAC_FORMAT_FLAG_8_BIT) ? 8 : ((header_old.nFormatFlags & MAC_FORMAT_FLAG_24_BIT) ? 24 : 16);
        samples = (header_old.nTotalFrames == 0) ? 0 : ((header_old.nTotalFrames -  1) * nBlocksPerFrame) + header_old.nFinalFrameBlocks;
        Info->Details.Duration      = (header_old.nSampleRate > 0)  ?  ((double)((signed __int64)samples) / (double)header_old.nSampleRate)  :  0.;
        Info->Details.Samples       = (__int64)samples;
        Info->Details.SampleRate    = (long)header_old.nSampleRate;
        Info->Details.Channels      = (long)header_old.nChannels;
        Info->Details.BitRate       = (Info->Details.Duration > 0)  ?  (long)((signed __int64)Info->FileSize * 8 / Info->Details.Duration + 500) / 1000  :  0;
        comp = compressionLevel( header_old.nCompressionLevel );
    }

    if ( header_all.nVersion % 10 == 0 ) {
        sprintf ( Info->Details.Format, "Monkey's Audio %u.%u (%s)", header_all.nVersion/1000, (header_all.nVersion%1000)/10, comp );
    } else {
        sprintf ( Info->Details.Format, "Monkey's Audio %u.%u (%s)", header_all.nVersion/1000, header_all.nVersion%1000, comp );
    }

    return 0;
}





